/*+++*
 *  title:	PackageInspector.m
 *  abstract:	NEXTSTEP Workspace Manager Inspector for Installer ".pkg" files.
 *  author:	T.R.Hageman, Groningen, The Netherlands
 *  created:	November 1994
 *  modified:	(see RCS Log at end)
 *  copyleft:
 *
 *		Copyright (C) 1994--1996,1999  Tom R. Hageman.
 *
 *	This is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This software is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this software; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *	02111-1307  USA.
 *
 *  description:
 *
 *---*/

#ifdef RCS_ID
static const char RCSid[] =
"PackageInspector.m,v 1.15 1999/08/24 21:06:41";
#endif

#define VERSION	"1.0"

#ifndef DEBUG
#   define DEBUG 0
#endif

#import "PackageInspector.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

// Localized strings
#define OPENBUTTON		NXLocalizedStringFromTableInBundle(NULL, bundle, "Open", NULL, okbutton label)
#define LISTCONTENTSBUTTON	NXLocalizedStringFromTableInBundle(NULL, bundle, "List Contents", NULL, revertbutton label)
#define LISTDESCRIPTIONBUTTON	NXLocalizedStringFromTableInBundle(NULL, bundle, "Description", NULL, revertbutton label)

// States
#define STATE_UNINSTALLED	NXLocalizedStringFromTableInBundle(NULL, bundle, "Uninstalled", NULL, original package state)
#define STATE_INSTALLED		NXLocalizedStringFromTableInBundle(NULL, bundle, "installed", "Installed", package has been uncompressed unto disk)
#define STATE_COMPRESSED	NXLocalizedStringFromTableInBundle(NULL, bundle, "compressed", "Compressed", installed package has been recompressed)

// so InfoView.strings can be ripped off from Installer.app
#define SIZEFORMAT		NXLocalizedStringFromTableInBundle("InfoView", bundle, "%s installed, %s compressed", NULL, Short indication to user about the size of a package once installed and the size when compressed)
#define KBYTES			NXLocalizedStringFromTableInBundle("InfoView", bundle, "KB", NULL, Kilobytes -- package size)
#define MBYTES			NXLocalizedStringFromTableInBundle("InfoView", bundle, "MB", NULL, MegaBytes -- package size)

#define SIZEGZIPFORMAT		NXLocalizedStringFromTableInBundle(NULL, bundle, "%s installed, %s gzipped", NULL, Short indication to user about the size of a package once installed and the size when compressed using gzip_package)

#define LOCALIZE(s)		NXLoadLocalized\
StringFromTableInBundle(NULL, bundle, s, NULL)
#define LOCALIZE_ARCH(s)	NXLoadLocalized\
StringFromTableInBundle("Architectures", bundle, s, NULL)


@implementation PackageInspector

- initWithNibName:(const char *)nibname
{
	char path[MAXPATHLEN+1];

	bundle = [NXBundle bundleForClass:[self class]];

	if (!([bundle getPath:path forResource:nibname ofType:"nib"] &&
	      [NXApp loadNibFile:path owner:self]))
	{
		fprintf(stderr, "%s: couldn't load %s.nib\n", [self name], nibname);
		[self free];
		self = nil;
	}
	return self;
}

+ new
{
	static PackageInspector *instance;
	
	if (instance == nil)
	{
		instance = [[super new] initWithNibName:[self name]];
	}
	return instance;
}

- awakeFromNib
{
	// NB. this is invoked after loading Info.nib as well(!)

	if (descriptionSwapView == nil)
	{
		NXSize maxSize;

		// Main nib.
		descriptionSwapView = [swapViewBox contentView];
		listContentsSwapView = [listContentsSwapView contentView];

		[packageDescriptionText setVertResizable:YES];

		[listContentsText setHorizResizable:YES];
		[listContentsText setVertResizable:YES];
		[listContentsText setCharWrap:NO];
		[listContentsText setNoWrap];
		[listContentsText getMaxSize:&maxSize];
		maxSize.width = 10000;
		[listContentsText setMaxSize:&maxSize];

		[inspectorVersionField setStringValue:VERSION];
	}
	[infoVersionField setStringValue:VERSION];

	return self;
}

- showInfo:sender
{
	if (infoPanel == nil)
	{
		char path[MAXPATHLEN+1];

		if ([bundle getPath:path forResource:"Info" ofType:"nib"])
		{
		    [NXApp loadNibFile:path owner:self];
		}
	}
	[infoPanel makeKeyAndOrderFront:sender];
	return self;
}

- revert:sender
{
	[super revert:sender];

	if ([self selectionCount] != 1)
	{
		return nil;
	}
	if (sender == [self revertButton])
	{
		[self toggleDescription];
	}
	else
	{
		char path[MAXPATHLEN+1];

		[package free];
		[self selectionPathsInto:path separator:'\0'];
		if (!(package = [[NXBundle allocFromZone:[self zone]] initForDirectory:path]))
		{
			return nil;
		}
		if ([self shouldLoad])
		{
			[self load];
			[self setRevertButtonState:listContents];
		}
	}
	[[[self okButton] setTitle:OPENBUTTON] setEnabled:YES];
	[self setRevertButtonTitle];

	return self;
}

- ok:sender
{
	[self perform:@selector(open:) with:sender afterDelay:0 cancelPrevious:NO];
	[super ok:sender];
	return self;
}

- load
{
	char buf[256], size[2][20];
	HashTable *table = [[HashTable alloc] initKeyDesc:"*" valueDesc:"*"];

	[self getArchs];

	// Collect information about the package in a hashtable.
	[self loadKeyValuesFrom:"info" inTable:table];
	[self loadKeyValuesFrom:"sizes" inTable:table];
	[self loadContentsOf:"location" inTable:table];
	[self loadContentsOf:"status" inTable:table];

	// Convenience macro.
#define LOOKUP(key, notfound)	([table isKey:key] ? [table valueForKey:key] : \
				 (notfound))
	[packageDescriptionText setText:LOOKUP("Description", "")];
	[packageLocationField setStringValue:
	 LOOKUP("location", LOOKUP("DefaultLocation", "???"))];

	[self formatSize:[table valueForKey:"InstalledSize"] inBuf:size[0]];
	[self formatSize:[table valueForKey:"CompressedSize"] inBuf:size[1]];
	// [TRH] this knows gzip_package internals... 
	sprintf(buf, [table isKey:"realCompressedSize"] ? SIZEGZIPFORMAT : SIZEFORMAT,
		size[0], size[1]);
	[packageSizesField setStringValue:buf];

	[packageStatusField setStringValue:LOCALIZE(LOOKUP("status", "Uninstalled"))];
	[packageTitleField setStringValue:LOOKUP("Title", "???")];
	[packageVersionField setStringValue:LOOKUP("Version", "???")];
#undef LOOKUP
	// Is this how one frees the contents of a hashtable?
	[table freeKeys:free values:free];
	[table free];

	[self loadImage];

	[self listContents];

	return self;
}

- loadKeyValuesFrom:(const char *)type inTable:(HashTable *)table
{
	char path[MAXPATHLEN+1];
	NXStream *stream;

	if (stream = NXMapFile([self getPath:path forType:type], NX_READONLY))
	{
		int c;

#if DEBUG & 1
		fprintf(stderr, "loadKeyValuesFrom:%s\n", path);
#endif
		while ((c = NXGetc(stream)) >= 0)
		{
			// Buffer sizes should be enough, according to doc.
			char key[1024+1], value[1024+1];
			char *p;

			if (NXIsSpace(c)) continue;
			if (c == '#')
			{
				while ((c = NXGetc(stream)) >= 0 && c != '\n') ;
				continue;
			}
			// Found key; collect it.
			p = key;
			do
			{
				if (p < &key[sizeof key-1]) *p++ = c;
			}
			while ((c = NXGetc(stream)) >= 0 && !NXIsSpace(c));
			*p = '\0';

			// Skip over spaces and tabs.
			while (c == ' ' || c == '\t') c = NXGetc(stream);

			// Value is rest of line, up to newline.
			p = value;
			do
			{
				if (p < &value[sizeof value-1]) *p++ = c;
			}
			while ((c = NXGetc(stream)) >= 0 && c != '\n');
			*p = '\0';

			// Insert key/value pair in hashtable.
#if DEBUG & 1
			fprintf(stderr, "key:%s value:%s\n", key, value);
#endif
			[table insertKey:NXCopyStringBuffer(key)
			 value:NXCopyStringBuffer(value)];
		}

		NXCloseMemory(stream, NX_FREEBUFFER);
	}
	return self;

}

- loadContentsOf:(const char *)type inTable:(HashTable *)table
{
	char path[MAXPATHLEN+1];
	NXStream *stream;

	if (stream = NXMapFile([self getPath:path forType:type], NX_READONLY))
	{
		char line[1024+1];
		int n = NXRead(stream, line, sizeof line);

		if (n > 0 && line[n-1] == '\n') line[n-1] = '\0';	// remove trailing newline.

		NXCloseMemory(stream, NX_FREEBUFFER);

		[table insertKey:NXCopyStringBuffer(type)
		 value:NXCopyStringBuffer(line)];
	}
	return self;
}

- loadImage
{
	char path[MAXPATHLEN+1];
	NXImage *image;

	// Remove old image from the button.
	if (image = [packageIconButton image])
	{
		[packageIconButton setImage:nil];
		[image free];
	}
	// Get the image (if any) from the package
	image = [[NXImage allocFromZone:[self zone]] initFromFile:[self getPath:path forType:"tiff"]];
	[packageIconButton setImage:image];

	return self;
}


#define STAT_EQ(s1, s2)	((s1)->st_ino == (s2)->st_ino && \
			 (s1)->st_dev == (s2)->st_dev && \
			 (s1)->st_mtime == (s2)->st_mtime && \
			 (s1)->st_size == (s2)->st_size)

- (BOOL)shouldLoad
{
	char path[MAXPATHLEN+1];
	struct stat newstats[NUMSTATS];
	static const char * const typesToStat[NUMSTATS] = { TYPESTOSTAT };
	BOOL result = NO;
	int i;

	for (i = 0;  i < NUMSTATS;  i++)
	{
		memset(&newstats[i], 0, sizeof(struct stat));
		if (!(stat([self getPath:path forType:typesToStat[i]], &newstats[i]) == 0 &&
		      STAT_EQ(&newstats[i], &stats[i])))
		{
			result = YES;
			///break; // NOT!!! must stat all for accurate cache.
		}
		stats[i] = newstats[i];
	}

	return result;
}

- toggleDescription
{
	switch (revertButtonState)
	{
	case listContents:
		[self setRevertButtonState:listDescription];
		break;
	case listDescription:
		[self setRevertButtonState:listContents];
		break;
	}
	return [self setRevertButtonTitle];
}

- (void)setRevertButtonState:(int)aValue
{
	if (revertButtonState != aValue)
	{
		revertButtonState = aValue;

		switch (revertButtonState)
		{
		case listContents:
			[swapViewBox setContentView:descriptionSwapView];
			break;
		case listDescription:
			[swapViewBox setContentView:listContentsSwapView];
			break;
		}
		[swapViewBox display];	// XXX Needed?
	}

}

// Support methods
- (const char *)getPath:(char *)buf forType:(const char *)type
{
	char name[MAXPATHLEN+1];

	// Get package name, sans extension.
	*strrchr(strcpy(name, strrchr([package directory], '/')+1), '.') = '\0';

	// Now get the full pathname.
	[package getPath:buf forResource:name ofType:type];
#if DEBUG & 2
	fprintf(stderr, "PackageInspector: type=\"%s\" name=\"%s\" path=\"%s\"\n",
		type, name, buf);
#endif
	return buf;
}

- setRevertButtonTitle
{
	[[[self revertButton]
	  setTitle:LOCALIZE(revertButtonState == listContents ?
			    "List Contents" : "Description")]
	 setEnabled:YES];
	return self;
}

- (const char *)formatSize:(const char *)size inBuf:(char *)buf
{
	// [TRH] this is very simplistic (but seems consistent with Installer.app)
	if (!size)
	{
		strcpy(buf, "???");
	}
	else
	{
		int len = strlen(size);
		if (len < 4)
		{
			sprintf(buf, "%s%s", size, KBYTES);
		}
		else if (len < 6)
		{
			sprintf(buf, "%.*s.%.*s%s",
				(len-3), size, 3-(len-3), size+(len-3), MBYTES); 
		}
		else
		{
			sprintf(buf, "%.*s%s", (len-3), size, MBYTES);
		}
	}
	return buf;
}

/* Create a subprocess, using `script' as the command name, to be found in the
   inspector's bundle, and the *.bom inside the package as its argument. */
- (Subprocess *)createSubprocess:(const char *)script
{
	char filename[MAXPATHLEN+1];
	char command[4*MAXPATHLEN+10+1];
	const char *s;
	char *d = command;

	for (s = "exec "; *s; ) *d++ = *s++;

	[bundle getPath:d forResource:script ofType:NULL];
	d += strlen(d);

	for (s = " '"; *s; ) *d++ = *s++;

	[self getPath:filename forType:"bom"];

	// Escape single quote characters.
	for (s = filename; *s; )
	{
		if ((*d++ = *s++) == '\'')
		{
			*d++ = '\\', *d++ = '\'', *d++ = '\'';
		}
	}
	for (s = "'"; *d++ = *s++; ) ;

	return [[Subprocess allocFromZone:[self zone]] init:command
			withDelegate:self andPtySupport:NO andStdErr:NO];

}

// Determine architectures, in separate subprocess.

#define WORKING	" ..."	// `I'm still busy' indicator.

- (void)getArchs
{
	if (archProcess) [archProcess terminate:self];

	[packageArchesField setStringValue:WORKING];

	archProcess = [self createSubprocess:"archbom"];
}

- (void)addArchs:(const char *)string
{
	char result[1024];	// Should be big enough...
	const char *s;
	char *d;

	strcpy(result, [packageArchesField stringValue]);
	if ((d = strstr(result, WORKING)) != NULL)
	{
		*d = '\0';
	}
	else
	{
		d = result + strlen(result);
	}
	if ((s = string))
	{
		do
		{
			char name[100];
			char *t = name;

			while (*s && !NXIsAlNum(*s))
			{
				if (*s == '\n')
				{
					*d++ = ' ', s++;
				}
				else
				{
					*d++ = *s++;
				}
			}
			while (NXIsAlNum(*s)) *t++ = *s++;
			*t = '\0';
			if (t > name)
			{
#if DEBUG & 4
				fprintf(stderr, "addArchs:\"%s\" localized: \"%s\"\n", name, LOCALIZE_ARCH(name));
#endif
				strcpy(d, LOCALIZE_ARCH(name));
				d += strlen(d);
			}
		}
		while (*s);

		strcpy(d, WORKING);
	}
	[packageArchesField setStringValue:result];
	[window displayIfNeeded]; // necessary??
}

// List contents.

- (void)listContents
{
	id listScrollView = [[listContentsText superview] superview];

	if (listProcess) [listProcess terminate:self];

	[listScrollView setCopyOnScroll:NO];
	// ...to workaround display bugs with horiz. scrolling of Text.
	[listScrollView setHorizScrollerRequired:YES];
	[listContentsText setText:""];

	listProcess = [self createSubprocess:"listbom"];
}

- (void)addListContents:(const char *)buffer
{
	int len = [listContentsText textLength];

	if (buffer)
	{
		[listContentsText setSel:len:len];
		[listContentsText replaceSel:buffer];
#if 0
		/* The following certainly looks better in its intermediate
		   stage for long lines, but also slows us down significantly... */
		[listContentsText calcLine];
		[listContentsText sizeToFit];
		[listContentsText display];
#endif
	}
	else
	{
		if (len > 0)
		{	// Strip off trailing newline.
			char c;

			[listContentsText getSubstring:&c start:(len-1) length:1];
			if (c == '\n')
			{
				[listContentsText setSel:(len-1):len];
				[listContentsText replaceSel:""];
			}
		}
		[listContentsText calcLine];
		[listContentsText sizeToFit];
		[listContentsText display];
		/* (There's still something whacky with this calcLine /
		   sizeToFit stuff: for (some?) long lines, the selectable area
		   is still multiple lines (ie wrapped) instead of a single
		   line, so it looks like something has not been updated
		   properly.  But I can't think of anything I can do anymore
		   to remedy this...
		   OK, the -setCharWrap:NO, -setNoWrap above seems to help.) */
	    {
		/* Remove horizontal scrollbar, if possible. */
		NXRect visibleRect, textBounds;

		[listContentsText getVisibleRect:&visibleRect];
		[listContentsText getBounds:&textBounds];

		if (visibleRect.size.width >= textBounds.size.width)
		{
			id listScrollView = [[listContentsText superview] superview];

			[listScrollView setCopyOnScroll:YES];
			[listScrollView setHorizScrollerRequired:NO];
		}
	    }
	}
}

// Subprocess delegate methods.

- subprocess:(Subprocess *)sender output:(char *)buffer
{
	if (sender == archProcess)
	{
		[self addArchs:buffer];
	}
	if (sender == listProcess)
	{
		[self addListContents:buffer];
	}
	return self;
}

- subprocessDone:(Subprocess *)sender
{
	[self subprocess:sender output:NULL];

	if (sender == archProcess)
	{
		archProcess = nil;
	}
	if (sender == listProcess)
	{
		listProcess = nil;
	}
	[sender free];
	return self;
}

static void openInWorkspace(const char *filename)
{
	// Indirect approach to circumvent Workspace deadlock/timeout.
	char command[14+3*MAXPATHLEN+1];
	const char *s;
	char *d = command;

	for (s = "exec open '"; *s; ) *d++ = *s++;
	// Escape single quote characters.
	for (s = filename; *s; )
	{
		if ((*d++ = *s++) == '\'')
		{
			*d++ = '\\', *d++ = '\'', *d++ = '\'';
		}
	}
	for (s = "'&"; *d++ = *s++; ) ;
	system(command);
}

- open:sender
{
	openInWorkspace([package directory]);
	return self;
}

@end // PackageInspector

/*======================================================================
 * PackageInspector.m,v
 * Revision 1.15  1999/08/24 21:06:41  tom
 * PackageInspector:
 * (#VERSION): set to 1.0.
 *
 * README.rtf:
 * - prepare for release 1.0.
 *
 * Revision 1.14  1999/08/15 17:14:03  tom
 * PackageInspector:
 * (#VERSION): set to 0.98.
 *
 * README:
 * - prepare for 0.98 [BETA] release.
 *
 * Revision 1.13  1999/08/15 15:56:21  tom
 * - localization support changes:
 * (#OPENBUTTON,#LISTCONTENTSBUTTON,#LISTDESCRIPTIONBUTTON): make comments more
 *  explicit.
 * (#LOCALIZE,#LOCALIZE_ARCH): newline-escape to avoid genstrings warnings.
 *
 * Revision 1.12  1999/08/14 14:48:55  tom
 * PackageInspector:
 * (//copyleft:): update FSF snail-mail address.
 * (#VERSION): set to 0.973
 * (-initWithNibName:): new method, factored out from +new.
 * (+new): use it.
 * (-awakeFromNib): initialize listContentsText to not line-wrap at all.
 * (-listContents): remove redundant listContentsText setup.
 * - general opening-brace reformatting (now on line of its own.)
 *
 * Revision 1.11  1999/08/04 14:22:32  tom
 * PackageInspector:
 * (-listContents): disable word- and character wrap to workaround apparent
 *  -calcLine/selection bug in Text view.
 *
 * Revision 1.10  1999/08/04 13:23:07  tom
 * PackageInspector:
 * - add "List Contents" support:
 * (listContentsText,swapViewBox,listContentsSwapView,descriptionSwapView):
 *  new outlets.
 * (listProcess): new ivar.
 * (-setRevertButtonState:,-listContents): new public methods.
 * (-awakeFromNib,-createSubprocess:,-addListContents:): new support methods.
 * (-revert:,-toggleDescription):  use -setRevertButtonState:.
 * (+new,-showInfo:): move outlet initializations to -awakeFromNib.
 * (-load): invoke -listContents.
 * (-getArchs): use -createSubprocess:.
 * (-subprocess:output:,-subprocessDone:): handle listProcess.
 *
 * README.rtf:
 * - udpdated for 0.97 BETA release.
 *
 * Revision 1.9  1996/03/18 22:02:53  tom
 * (VERSION): 0.96 -- prepare for release;
 * (-load): check for gzip_package'd package, use appropriate sizes format string.
 *
 * Revision 1.8  1995/09/01 21:46:27  tom
 * Circumvent open deadlock/timeout (when Installer.app is not yet launched);
 * (openInWorkspace): new private function; (-open:): new method.
 *
 * Revision 1.7  1995/07/30 22:20:26  tom
 * (LOCALIZE_ARCH): new macro; (-addArchs:): new method;
 * (-subprocess:output:,-subprocessDone:) use it.
 *
 * Revision 1.6  1995/07/30 16:59:51  tom
 * import Subprocess.h; (archProcess): new ivar;
 * (-getArchs,-subprocess:output:,-subprocessDone:): new methods;
 * added for asynchronous arch-determination.
 *
 * Revision 1.5  1995/07/29 19:13:35  tom
 * (+new): avoid reassignment of self;
 *  make packageDescriptionText vertically resizable;
 * (-shouldLoad): rewritten to generalized array-driven approach.
 *
 * Revision 1.4  1995/04/02 02:39:01  tom
 * (package): NXBundle instead of (const char *). so that localized info files
 *  are found. (this loses out if *.pkg is a symbolic link, though.)
 *
 * Revision 1.3  1994/12/07 00:00:36  tom
 * (RCSid): add spaces.
 *
 * Revision 1.2  1994/11/25 21:27:18  tom
 * (package ivar): use (char*) instead of (NXBundle*) to workaround symlink problems
 *
 * Revision 1.1  1994/11/25 16:13:12  tom
 * Initial revision
 *
 *======================================================================*/
